# pylint: disable=duplicate-code


import numpy as np

import bencher as bch


class VolumeSample(bch.ParametrizedSweep):
    """A class to represent a 3D point in space."""

    x = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="x coordinate of the sample volume", samples=9
    )
    y = bch.FloatSweep(
        default=0, bounds=[-2.0, 2.0], doc="y coordinate of the sample volume", samples=10
    )
    z = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="z coordinate of the sample volume", samples=11
    )

    surf_x = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="z coordinate of the sample volume", samples=9
    )

    surf_y = bch.FloatSweep(
        default=0, bounds=[-1.0, 1.0], doc="z coordinate of the sample volume", samples=11
    )


class VolumeResult(bch.ParametrizedSweep):
    """A class to represent the properties of a volume sample."""

    # value = bch.ResultVar("ul", doc="The scalar value of the 3D volume field")
    p1_dis = bch.ResultVar("m", direction=bch.OptDir.minimize, doc="The distance to p1")
    p2_dis = bch.ResultVar("m", direction=bch.OptDir.minimize, doc="The distance to p2")
    total_dis = bch.ResultVar(
        "m", direction=bch.OptDir.minimize, doc="The total distance to all points"
    )
    surf_value = bch.ResultVar(
        "ul", direction=bch.OptDir.maximize, doc="The scalar value of the 3D volume field"
    )
    # occupancy = bch.ResultVar("occupied", doc="If the value is > 0.5 this point is considered occupied")


p1 = np.array([0.4, 0.6, 0.0])
p2 = np.array([-0.8, -0.2, 0.0])
surf_x_max = 0.5
surf_y_max = -0.2


def bench_fn(point: VolumeSample) -> VolumeResult:
    """This function takes a 3D point as input and returns distance of that point to the origin.

    Args:
        point (VolumeSample): Sample point

    Returns:
        VolumeResult: Value at that point
    """

    cur_point = np.array([point.x, point.y, point.z])

    output = VolumeResult()
    output.p1_dis = np.linalg.norm(p1 - cur_point)
    output.p2_dis = np.linalg.norm(p2 - cur_point)
    output.total_dis = output.p1_dis + output.p2_dis

    output.surf_value = output.total_dis + np.linalg.norm(
        np.array(
            [np.cos(point.surf_x - surf_x_max), 2 * np.cos(point.surf_y - surf_y_max), point.z]
        )
    )

    return output


def example_floats2D_workflow(run_cfg: bch.BenchRunCfg, bench: bch.Bench = None) -> bch.Bench:
    """Example of how to perform a 3D floating point parameter sweep

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """
    if bench is None:
        bench = bch.Bench("Bencher_Example_Floats", bench_fn, VolumeSample)

    # run_cfg.debug = False

    res = bench.plot_sweep(
        input_vars=[VolumeSample.param.x, VolumeSample.param.y],
        result_vars=[
            VolumeResult.param.total_dis,
            VolumeResult.param.p1_dis,
            VolumeResult.param.p2_dis,
        ],
        const_vars=[VolumeSample.param.z.with_const(0)],
        title="Float 2D Example",
        run_cfg=run_cfg,
    )
    recovered_p1 = res.get_optimal_vec(VolumeResult.param.p1_dis, res.input_vars)
    print(f"recovered p1: {recovered_p1}, distance: {np.linalg.norm(recovered_p1 - p1[:2])}")
    # within tolerance of sampling
    # assert np.linalg.norm(recovered_p1 - p1[:2]) < 0.15

    recovered_p2 = res.get_optimal_vec(VolumeResult.param.p2_dis, res.input_vars)
    print(f"recovered p2: {recovered_p2} distance: {np.linalg.norm(recovered_p2 - p2[:2])}")
    # within tolerance of sampling
    # assert np.linalg.norm(recovered_p2 - p2[:2]) < 0.15

    run_cfg.use_optuna = True
    for rv in res.result_vars:
        bench.plot_sweep(
            input_vars=[VolumeSample.param.surf_x, VolumeSample.param.surf_y],
            result_vars=[
                VolumeResult.param.surf_value,
            ],
            const_vars=res.get_optimal_inputs(rv, True),
            title=f"Slice of 5D space for {rv.name}",
            description="""This example shows how to sample 3 floating point variables and plot a 3D vector field of the results.""",
            run_cfg=run_cfg,
        )

    return bench


def example_floats3D_workflow(run_cfg: bch.BenchRunCfg, bench: bch.Bench = None) -> bch.Bench:
    """Example of how to perform a 3D floating point parameter sweep

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """
    if bench is None:
        bench = bch.Bench("Bencher_Example_Floats", bench_fn, VolumeSample)

    # run_cfg.debug = False

    res = bench.plot_sweep(
        input_vars=[VolumeSample.param.x, VolumeSample.param.y, VolumeSample.param.z],
        result_vars=[
            VolumeResult.param.total_dis,
            VolumeResult.param.p1_dis,
            VolumeResult.param.p2_dis,
            # get_optimal_inputsm.occupancy,
        ],
        title="Float 3D Example",
        description="""This example shows how to sample 3 floating point variables and plot a volumetric representation of the results.  The benchmark function returns the distance to the origin""",
        post_description="Here you can see concentric shells as the value of the function increases with distance from the origin. The occupancy graph should show a sphere with radius=0.5",
        run_cfg=run_cfg,
    )

    recovered_p1 = res.get_optimal_vec(VolumeResult.param.p1_dis, res.input_vars)
    print(f"recovered p1: {recovered_p1}, distance: {np.linalg.norm(recovered_p1 - p1)}")
    # within tolerance of sampling
    # assert np.linalg.norm(recovered_p1 - p1) < 0.15

    recovered_p2 = res.get_optimal_vec(VolumeResult.param.p2_dis, res.input_vars)
    print(f"recovered p2: {recovered_p2} distance: {np.linalg.norm(recovered_p2 - p2)}")
    # within tolerance of sampling
    # assert np.linalg.norm(recovered_p2 - p2) < 0.15

    run_cfg.use_optuna = True
    for rv in res.result_vars:
        bench.plot_sweep(
            input_vars=[VolumeSample.param.surf_x, VolumeSample.param.surf_y],
            result_vars=[
                VolumeResult.param.surf_value,
            ],
            const_vars=res.get_optimal_inputs(rv, True),
            title=f"Slice of 5D space for {rv.name}",
            description="""This example shows how to sample 3 floating point variables and plot a 3D vector field of the results.""",
            run_cfg=run_cfg,
        )

    return bench


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()

    bench_ex = example_floats2D_workflow(ex_run_cfg)
    example_floats3D_workflow(ex_run_cfg, bench_ex)

    bench_ex.show()



import unittest
import bencher as bch
from bencher.example.example_sample_cache import UnreliableClass
from bencher.example.example_sample_cache_context import example_cache_context


class TestSampleCache(unittest.TestCase):
    def call_bencher(self, bencher, run_cfg):
        bencher.plot_sweep(
            title="Example Crashy Function with the sample_cache",
            input_vars=[UnreliableClass.param.input_val],
            result_vars=[
                UnreliableClass.param.return_value,
                UnreliableClass.param.trigger_crash,
            ],
            description="""This example shows how to use the use_sample_cache option to deal with unreliable functions and to continue benchmarking using previously calculated results even if the code crashing during the run""",
            run_cfg=run_cfg,
            post_description="The input_val vs return value graph is a straigh line as expected and there is no record of the fact the benchmark crashed halfway through. The second graph shows that for values >1 the trigger_crash value had to be 0 in order to proceed",
            tag="testing_tag3",
        )

    def test_sample_cache(self):
        """The sample cache function needs to be run twice because the first run can pass when there is no cache, but will fail the second time when the cache exists"""
        self.sample_cache()
        self.sample_cache()

    def sample_cache(self):
        run_cfg = bch.BenchRunCfg()
        run_cfg.repeats = 1

        run_cfg.use_sample_cache = True  # this will store the result of every call
        run_cfg.only_hash_tag = True

        instance = UnreliableClass()
        instance.trigger_crash = True

        bencher = bch.Bench("example_sample_cache", instance.crashy_fn)
        bencher.clear_tag_from_cache("testing_tag3")

        # the benchmark is set up to clear the previous sample cache and to cache the intermediate results from each benchmark sample.  It will throw an exception because the class has been set up to crash after the 2nd sample
        with self.assertRaises(RuntimeError):
            self.call_bencher(bencher, run_cfg)

        # worker is called 3 times (and crashes on the 3rd), there should be no values in the cache
        self.assertEqual(bencher.worker_wrapper_call_count, 3)
        self.assertEqual(bencher.worker_fn_call_count, 3)
        self.assertEqual(bencher.worker_cache_call_count, 0)

        # now turn off the clearing of the sample cache.  Previously calculated values will be used now

        # turn off the trigger to crash so the remaining values can be calculated
        instance.trigger_crash = False

        # clear the call counts from the previous run
        bencher.clear_call_counts()
        # this will calculate the remaining values
        self.call_bencher(bencher, run_cfg)

        # the wrapped call count should be 4 because there are 4 samples
        self.assertEqual(bencher.worker_wrapper_call_count, 4)
        # the fn_call count should be 2 because 2 values were left uncalculated
        self.assertEqual(bencher.worker_fn_call_count, 2)
        # the cache call count should be 2 because 2 values were successfully calculated last time (3rd one crashed before the result was stored)
        self.assertEqual(bencher.worker_cache_call_count, 2)

        # now increase the number of repeats to 2.  This will use the previously calculated values for repeat 1 and only calculate the new values for repeat=2
        run_cfg.repeats = 2
        bencher.clear_call_counts()
        self.call_bencher(bencher, run_cfg)

        # 4 samples x 2 repeats = 8 calls
        self.assertEqual(bencher.worker_wrapper_call_count, 8)
        # there should be 4 new calls for each point of the sweep for the second repeat
        self.assertEqual(bencher.worker_fn_call_count, 4)
        # the first 4 values from repeat=1 are loaded from the cache
        self.assertEqual(bencher.worker_cache_call_count, 4)

        # create a new benchmark and clear the cache (for this new benchmark only)
        run_cfg.clear_sample_cache = True
        bencher.clear_call_counts()
        bencher.plot_sweep(
            title="new benchmark", result_vars=[UnreliableClass.param.return_value], run_cfg=run_cfg
        )

        # 1 sample x 2 repeats
        self.assertEqual(bencher.worker_wrapper_call_count, 2)
        # 1 sample x 2 repeats which are not in the cache because this is a different benchmark
        self.assertEqual(bencher.worker_fn_call_count, 2)
        # sample cache is cleared so no values from the cache
        self.assertEqual(bencher.worker_cache_call_count, 0)

        # check that the previous benchmark still has the cached values, and that they were not removed by clearing the sample cache of an unrelated benchmark

        run_cfg.clear_sample_cache = False
        bencher.clear_call_counts()
        self.call_bencher(bencher, run_cfg)

        # 4 samples x 2 repeats
        self.assertEqual(bencher.worker_wrapper_call_count, 8)
        # all the results should be in the cache and not been cleared by the previous clear_sample_cache=True call because this is a different benchmark
        self.assertEqual(bencher.worker_fn_call_count, 0)
        # all 8 previously calculated results already in the cache
        self.assertEqual(bencher.worker_cache_call_count, 8)

    def test_sample_cache_context(self):
        """The sample cache function needs to be run twice because the first run can pass when there is no cache, but will fail the second time when the cache exists"""
        example_cache_context()
        example_cache_context()


# TestSampleCache().sample_cache()



import bencher as bch


class UnreliableClass(bch.ParametrizedSweep):
    """This class helps demonstrate benchmarking a function that sometimes crashes during sampling.  By using BenchRunCfg.use_sample_cache you can store the results of every call to the benchmark function so data is not lost in the event of a crash.  However, because cache invalidation is hard (https://martinfowler.com/bliki/TwoHardThings.html) you need to be mindful of how you could get bad results due to incorrect cache data.  For example if you change your benchmark function and use the sample cache you will not get correct values; you will need to use BenchRunCfg.clear_sample_cache to purge any out of date results."""

    input_val = bch.IntSweep(
        default=0,
        bounds=[0, 3],
        doc="If check limit=True the crashy_fn will crash if this value is >1",
    )
    return_value = bch.ResultVar(
        units="ul",
        doc="This is a dummy result variable. In this example, it is the same as the value passed in.",
    )
    trigger_crash = bch.ResultVar(
        units="True/False",
        doc="if true crashy_fn will crash when input_val >1",
    )

    def crashy_fn(self, input_val: int = 0, **kwargs) -> float:  # pylint: disable=unused-argument
        if self.trigger_crash:
            if input_val > 1:
                raise RuntimeError("I crashed for no good reason ;P")

        return {"return_value": input_val, "trigger_crash": self.trigger_crash}


def example_sample_cache(run_cfg: bch.BenchRunCfg, trigger_crash: bool) -> bch.Bench:
    """This example shows how to use the use_sample_cache option to deal with unreliable functions and to continue benchmarking using previously calculated results even if the code crashed during the run

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep
        trigger_crash: (bool): Turn on/off code to artificially trigger a crash

    Returns:
        Bench: results of the parameter sweep
    """

    instance = UnreliableClass()
    instance.trigger_crash = trigger_crash

    bencher = bch.Bench("example_sample_cache", instance.crashy_fn)

    bencher.plot_sweep(
        title="Example Crashy Function with the sample_cache",
        input_vars=[UnreliableClass.param.input_val],
        result_vars=[UnreliableClass.param.return_value, UnreliableClass.param.trigger_crash],
        description="""This example shows how to use the use_sample_cache option to deal with unreliable functions and to continue benchmarking using previously calculated results even if the code crashed during the run""",
        run_cfg=run_cfg,
        post_description="The input_val vs return value graph is a straight line as expected and there is no record of the fact the benchmark crashed halfway through. The second graph shows that for values >1 the trigger_crash value had to be 0 in order to proceed",
    )
    return bencher


if __name__ == "__main__":
    ex_run_cfg = bch.BenchRunCfg()
    ex_run_cfg.repeats = 1

    # this will store the result of of every call to crashy_fn
    ex_run_cfg.use_sample_cache = True
    ex_run_cfg.clear_sample_cache = True

    try:
        # this will crash after iteration 2 because we are checking the crash_threshold >1.  We don't want to lose those (potentially expensive to calculate) datapoints so they are stored in the sample_cache
        example_sample_cache(ex_run_cfg, trigger_crash=True)
    except RuntimeError as e:
        print(f"caught the exception {e}")

    print(
        "Running the same benchmark but without checking the limit.  The benchmarking should load the previously calculated values and continue to finish calculating the values that were missed due to the crash"
    )
    ex_run_cfg.clear_sample_cache = False
    example_sample_cache(ex_run_cfg, trigger_crash=False)

    ex_run_cfg.repeats = 2

    example_sample_cache(ex_run_cfg, trigger_crash=False).show()

    # see the test_sample_cache for a more detailed explanation of the mechanisms of the cache



# pylint: disable=duplicate-code
import bencher as bch

# All the examples will be using the data structures and benchmark function defined in this file
from bencher.example.benchmark_data import (
    ExampleBenchCfgIn,
    ExampleBenchCfgOut,
    NoiseDistribution,
    bench_function,
)


def example_floats2D(run_cfg: bch.BenchRunCfg) -> bch.Bench:
    """Example of how to perform a 2D floating point parameter sweep

    Args:
        run_cfg (BenchRunCfg): configuration of how to perform the param sweep

    Returns:
        Bench: results of the parameter sweep
    """
    bench = bch.Bench(
        "Bencher_Example_Floats",
        bench_function,
        ExampleBenchCfgIn,
    )

    cfg = ExampleBenchCfgIn()
    cfg.param.theta.samples = 3
    cfg.param.offset.samples = 2
    # TODO Create a test to confirm behavior

    run_cfg.use_optuna = True

    bench.plot_sweep(
        input_vars=[cfg.param.theta, cfg.param.offset],
        result_vars=[ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_cos],
        const_vars=[
            cfg.param.sigma.with_const(0.1),
            cfg.param.noise_distribution.with_const(NoiseDistribution.gaussian),
            cfg.param.noisy.with_const(True),
        ],
        title="Float 2D Example",
        description="""Bencher is a tool to make it easy to explore how input parameter affect a range of output metrics.  In these examples we are going to benchmark an example function which has been selected to show the features of bencher.
        The example function takes an input theta and returns the absolute value of sin(theta) and cos(theta) +- various types of noise.

        def bench_function(cfg: ExampleBenchCfgIn) -> ExampleBenchCfgOut:
            "Takes an ExampleBenchCfgIn and returns a ExampleBenchCfgOut output"
            out = ExampleBenchCfgOut()
            noise = calculate_noise(cfg)
            offset = 0.0

            postprocess_fn = abs if cfg.postprocess_fn == PostprocessFn.absolute else negate_fn

            out.out_sin = postprocess_fn(offset + math.sin(cfg.theta) + noise)
            out.out_cos = postprocess_fn(offset + math.cos(cfg.theta) + noise)
            return out

    The following examples will show how to perform parameter sweeps to characterise the behavior of the function.  The idea is that the benchmarking can be used to gain understanding of an unknown function.
        """,
        post_description="Here you can see the output plot of sin theta between 0 and pi.  In the tabs at the top you can also view 3 tabular representations of the data",
        run_cfg=run_cfg,
    )

    bench.plot_sweep(
        input_vars=[cfg.param.theta, cfg.param.offset, cfg.param.postprocess_fn],
        result_vars=[ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_cos],
        const_vars=[
            (cfg.param.sigma, 0.1),
            (cfg.param.noise_distribution, NoiseDistribution.gaussian),
            (cfg.param.noisy, True),
        ],
        title="Float 2D with categorical Example",
        description="""Here we add plot a 2d surface and facet over a categorical variable
        """,
        run_cfg=run_cfg,
    )

    bench.plot_sweep(
        input_vars=[
            cfg.param.theta,
            cfg.param.offset,
            cfg.param.postprocess_fn,
            cfg.param.noise_distribution,
        ],
        result_vars=[ExampleBenchCfgOut.param.out_sin, ExampleBenchCfgOut.param.out_cos],
        const_vars=[
            (cfg.param.sigma, 0.1),
            (cfg.param.noise_distribution, NoiseDistribution.gaussian),
            (cfg.param.noisy, True),
        ],
        title="Float 2D with categorical x2 Example",
        description="""Here we add plot a 2d surface and facet over two categorical variable
        """,
        run_cfg=run_cfg,
    )

    return bench


if __name__ == "__main__":
    example_floats2D(bch.BenchRunCfg(repeats=2)).show()
